
from typing import cast

from logging import Logger
from logging import getLogger

from untangle import Element

from pyutmodelv2.PyutUseCase import PyutUseCase

from umlshapes.shapes.UmlUseCase import UmlUseCase

from umlio.IOTypes import Elements
from umlio.IOTypes import GraphicInformation
from umlio.IOTypes import UmlUseCases
from umlio.IOTypes import umlUseCasesFactory

from umlio.XMLConstants import XmlConstants

from umlio.deserializer.XmlToPyut import XmlToPyut


class XmlUseCasesToUmlUseCases:
    """
    XML Elements to UML use case shape objects and associated model objects
    """
    def __init__(self):
        self.logger: Logger = getLogger(__name__)

        self._xmlToPyut: XmlToPyut = XmlToPyut()

    def deserialize(self, umlDiagramElement: Element) -> UmlUseCases:

        """

        Args:
            umlDiagramElement:  The Element document

        Returns:  deserialized UmlUseCase objects if any exist, else an empty list
        """
        umlUseCases:     UmlUseCases = umlUseCasesFactory()
        useCaseElements: Elements = cast(Elements, umlDiagramElement.get_elements(XmlConstants.ELEMENT_UML_USE_CASE))

        for useCaseElement in useCaseElements:
            self.logger.debug(f'{useCaseElement}')

            graphicInformation: GraphicInformation = GraphicInformation.toGraphicInfo(graphicElement=useCaseElement)
            pyutUseCase:        PyutUseCase        = self._xmlToPyut.useCaseToPyutUseCase(umlUseCaseElement=useCaseElement)
            umlUseCase:         UmlUseCase         = UmlUseCase(pyutUseCase=pyutUseCase)

            umlUseCase.id       = graphicInformation.id
            umlUseCase.size     = graphicInformation.size
            umlUseCase.position = graphicInformation.position

            umlUseCases.append(umlUseCase)

        return umlUseCases




from logging import Logger
from logging import getLogger

from xml.etree.ElementTree import Element
from xml.etree.ElementTree import SubElement

from umlshapes.links.UmlLollipopInterface import UmlLollipopInterface

from umlio.IOTypes import ElementAttributes
from umlio.IOTypes import UmlLollipopInterfaces
from umlio.XMLConstants import XmlConstants
from umlio.serializer.PyutToXml import PyutToXml


class UmlLollipopsToXmlLollipops:
    def __init__(self):

        self.logger: Logger = getLogger(__name__)

        self._pyutToXml: PyutToXml = PyutToXml()

    def serialize(self, documentTop: Element, umlLollipops: UmlLollipopInterfaces) -> Element:

        for umlLollipop in umlLollipops:
            self._umlLollipopInterfaceToXml(documentElement=documentTop, umlLollipopInterface=umlLollipop)

        return documentTop

    def _umlLollipopInterfaceToXml(self, documentElement: Element, umlLollipopInterface: UmlLollipopInterface) -> Element:
        """

        Args:
            documentElement:        Xml Element
            umlLollipopInterface:   Lollipop to serialize

        Returns: A new Element
        """
        attachedToId: str = umlLollipopInterface.attachedTo.id
        attributes: ElementAttributes = ElementAttributes({
            XmlConstants.ATTRIBUTE_LINE_CENTUM:     str(umlLollipopInterface.lineCentum),
            XmlConstants.ATTRIBUTE_ATTACHMENT_SIDE: umlLollipopInterface.attachmentSide.value,
            XmlConstants.ATTRIBUTE_ATTACHED_TO_ID:  attachedToId,
        })
        lollipopElement: Element = SubElement(documentElement, XmlConstants.ELEMENT_LOLLIPOP, attrib=attributes)

        self._pyutToXml.pyutInterfaceToXml(umlLollipopInterface.pyutInterface, lollipopElement)

        return lollipopElement




from typing import Dict
from typing import List
from typing import NewType
from typing import Union
from typing import cast

from logging import Logger
from logging import getLogger

from dataclasses import dataclass

from wx import Point

from untangle import Element

from codeallybasic.SecureConversions import SecureConversions

from pyutmodelv2.PyutLink import LinkDestination
from pyutmodelv2.PyutLink import LinkSource
from pyutmodelv2.PyutLink import PyutLink
from pyutmodelv2.enumerations.PyutLinkType import PyutLinkType

from umlshapes.types.Common import EndPoints

from umlshapes.links.UmlLink import UmlLink
from umlshapes.links.UmlNoteLink import UmlNoteLink
from umlshapes.links.UmlAssociation import UmlAssociation
from umlshapes.links.UmlInheritance import UmlInheritance
from umlshapes.links.UmlComposition import UmlComposition
from umlshapes.links.UmlAggregation import UmlAggregation

from umlshapes.shapes.UmlClass import UmlClass
from umlshapes.shapes.UmlNote import UmlNote
from umlshapes.shapes.UmlText import UmlText
from umlshapes.shapes.UmlUseCase import UmlUseCase
from umlshapes.shapes.UmlActor import UmlActor


from umlshapes.ShapeTypes import LinkableUmlShape
from umlshapes.ShapeTypes import LinkableUmlShapes

from umlio.IOTypes import Elements
from umlio.IOTypes import UmlLinkAttributes
from umlio.IOTypes import UmlLinks
from umlio.IOTypes import umlLinksFactory

from umlio.XMLConstants import XmlConstants

from umlio.deserializer.XmlToPyut import XmlToPyut


@dataclass
class ConnectedShapes:
    sourceShape:        LinkableUmlShape
    destinationShape:   LinkableUmlShape


UmlAssociationClasses = Union[UmlAssociation, UmlComposition, UmlAggregation]

# AssociationClassDescriptor = type[UmlAssociationClasses] = cast(type[LinkEventHandler], None)

LinkTypeToClass = NewType('LinkTypeToClass', Dict[PyutLinkType, type[UmlAssociationClasses]])

CLASSMAP: LinkTypeToClass = LinkTypeToClass(
    {
        PyutLinkType.ASSOCIATION: UmlAssociation,
        PyutLinkType.AGGREGATION: UmlAggregation,
        PyutLinkType.COMPOSITION: UmlComposition
    }
)
ASSOCIATION_LINK_TYPES: List[PyutLinkType] = [
    PyutLinkType.COMPOSITION, PyutLinkType.AGGREGATION, PyutLinkType.ASSOCIATION
]


class XmlLinksToUmlLinks:
    def __init__(self):
        self.logger: Logger = getLogger(__name__)

        self._xmlToPyut: XmlToPyut = XmlToPyut()

    def deserialize(self, umlDiagramElement: Element, linkableUmlShapes: LinkableUmlShapes) -> UmlLinks:

        umlLinks:     UmlLinks = umlLinksFactory()
        linkElements: Elements = cast(Elements, umlDiagramElement.get_elements(XmlConstants.ELEMENT_UML_LINK))

        for linkElement in linkElements:
            self.logger.debug(f'{linkElement=}')
            umlLink: UmlLink = self._umlLinkElementToUmlLink(umlLinkElement=linkElement, linkableUmlShapes=linkableUmlShapes)

            umlLinks.append(umlLink)

        return umlLinks

    def _umlLinkElementToUmlLink(self, umlLinkElement: Element, linkableUmlShapes: LinkableUmlShapes) -> UmlLink:

        pyutLinkElements: Elements = cast(Elements, umlLinkElement.get_elements(XmlConstants.ELEMENT_MODEL_LINK))
        assert len(pyutLinkElements) == 1, 'There can only be one'

        singlePyutLinkElement: Element = pyutLinkElements[0]        # I hate this short cut

        umlLink: UmlLink = self._getUmlLink(umlLinkElement=umlLinkElement,
                                            singlePyutLinkElement=singlePyutLinkElement,
                                            linkableUmlShapes=linkableUmlShapes
                                            )

        return umlLink

    def _getUmlLink(self, umlLinkElement: Element, singlePyutLinkElement: Element, linkableUmlShapes: LinkableUmlShapes) -> UmlLink:

        connectedShapes: ConnectedShapes = self._getConnectedShapes(singlePyutLinkElement, linkableUmlShapes)
        pyutLink:        PyutLink        = self._getPyutLink(singlePyutLinkElement, connectedShapes)

        umlLink: UmlLink = self._umlLinkFactory(srcShape=connectedShapes.sourceShape,
                                                pyutLink=pyutLink,
                                                destShape=connectedShapes.destinationShape,
                                                )

        umlLinkAttributes: UmlLinkAttributes = UmlLinkAttributes.fromGraphicLink(linkElement=umlLinkElement)
        self.logger.debug(f'{umlLinkAttributes}=')

        umlLink.id        = umlLinkAttributes.id
        umlLink.spline    = umlLinkAttributes.spline

        umlLink.MakeLineControlPoints(n=2)       # Make this configurable

        umlLink.endPoints = EndPoints(
            toPosition=umlLinkAttributes.toPosition,
            fromPosition=umlLinkAttributes.fromPosition
        )
        controlPoints: List[Point] = self._getLineControlPoints(umlLinkElement=umlLinkElement)
        for cp in controlPoints:
            umlLink.InsertLineControlPoint(point=cp)

        return umlLink

    def _getPyutLink(self, pyutLinkElement: Element, connectedShapes: ConnectedShapes) -> PyutLink:
        """

        Args:
            pyutLinkElement:    The Xml Elements
            connectedShapes:    The shapes at the ends of the link

        Returns:    A data model link
        """

        # noinspection PyUnresolvedReferences
        pyutLink: PyutLink = self._xmlToPyut.linkToPyutLink(
            singleLink=pyutLinkElement,
            source=self._getLinkSourceModelClass(connectedShapes.sourceShape),
            destination=self._getLinkDestinationModelClass(connectedShapes.destinationShape)
        )
        self.logger.debug(f'{pyutLink=}')

        return pyutLink

    def _getConnectedShapes(self, pyutLinkElement: Element, linkableUmlShapes: LinkableUmlShapes) -> ConnectedShapes:
        """

        Args:
            pyutLinkElement:
            linkableUmlShapes:   The dictionary of potential shapes

        Returns:  The connected shapes;  Will assert if it cannot find them
        """
        sourceId: int = int(pyutLinkElement[XmlConstants.ATTRIBUTE_SOURCE_ID])
        dstId:    int = int(pyutLinkElement[XmlConstants.ATTRIBUTE_DESTINATION_ID])

        try:
            sourceShape:      LinkableUmlShape = linkableUmlShapes[sourceId]
            destinationShape: LinkableUmlShape = linkableUmlShapes[dstId]
        except KeyError as ke:
            self.logger.error(f'{linkableUmlShapes=}')
            self.logger.error(f'Developer Error -- {pyutLinkElement=}')
            self.logger.error(f'Developer Error -- {sourceId=} {dstId=}  KeyError index: {ke}')
            assert False, 'Developer error'

        return ConnectedShapes(sourceShape=sourceShape, destinationShape=destinationShape)

    def _getLinkSourceModelClass(self, linkableUmlShape: LinkableUmlShape) -> LinkSource:
        """

        Args:
            linkableUmlShape:

        Returns:  The appropriate model class instance
        """

        if isinstance(linkableUmlShape, UmlClass):
            return linkableUmlShape.pyutClass
        elif isinstance(linkableUmlShape, UmlNote):
            return linkableUmlShape.pyutNote
        elif isinstance(linkableUmlShape, UmlText):
            return linkableUmlShape.pyutText
        elif isinstance(linkableUmlShape, UmlActor):
            return linkableUmlShape.pyutActor
        else:
            assert False, f'{linkableUmlShape=} is not a source linkable UML Shape'

    def _getLinkDestinationModelClass(self, linkableUmlShape: LinkableUmlShape) -> LinkDestination:
        """

        Args:
            linkableUmlShape:

        Returns: The appropriate model class instance
        """

        if isinstance(linkableUmlShape, UmlClass):
            return linkableUmlShape.pyutClass
        elif isinstance(linkableUmlShape, UmlUseCase):
            return linkableUmlShape.pyutUseCase
        else:
            assert False, f'{linkableUmlShape=} is not a destination linkable UML Shape'

    def _getLineControlPoints(self, umlLinkElement: Element) -> List[Point]:
        """
         <LineControlPoint x="100" y="100" />

        Args:
            umlLinkElement:

        Returns:
        """

        controlPoints: List[Point] = []

        controlPointElements: Elements = cast(Elements, umlLinkElement.get_elements(XmlConstants.ELEMENT_MODEL_LINE_CONTROL_POINT))
        for controlPointElement in controlPointElements:
            x: int = SecureConversions.secureInteger(controlPointElement[XmlConstants.ATTRIBUTE_X])
            y: int = SecureConversions.secureInteger(controlPointElement[XmlConstants.ATTRIBUTE_Y])

            point: Point = Point(x=x, y=y)
            controlPoints.append(point)

        return controlPoints

    def _umlLinkFactory(self, srcShape: LinkableUmlShape, pyutLink: PyutLink, destShape: LinkableUmlShape) -> UmlLink:
        """

        Args:
            srcShape:
            pyutLink:
            destShape:

        Returns:  The appropriate UML Link shape

        """
        if pyutLink.linkType == PyutLinkType.INHERITANCE:
            # Note dest and source are reversed here
            return UmlInheritance(baseClass=cast(UmlClass, destShape), pyutLink=pyutLink, subClass=cast(UmlClass, srcShape))
        elif pyutLink.linkType in ASSOCIATION_LINK_TYPES:
            umlAssociation = CLASSMAP[pyutLink.linkType](pyutLink)
            #
            # Need to do this because the shape is not yet on a canvas
            #
            umlAssociation.sourceShape      = srcShape
            umlAssociation.destinationShape = destShape
            return umlAssociation
        elif pyutLink.linkType == PyutLinkType.NOTELINK:
            umlNoteLink: UmlNoteLink = UmlNoteLink(pyutLink=pyutLink)
            umlNoteLink.sourceNote       = cast(UmlNote, srcShape)
            umlNoteLink.destinationClass = cast(UmlClass, destShape)
            return umlNoteLink
        else:
            assert False, f'Unknown link type, {pyutLink.linkType=}'




from typing import cast

from logging import Logger
from logging import getLogger

from untangle import Element

from pyutmodelv2.PyutNote import PyutNote
from umlshapes.shapes.UmlNote import UmlNote

from umlio.IOTypes import Elements
from umlio.IOTypes import GraphicInformation
from umlio.IOTypes import UmlNotes
from umlio.IOTypes import umlNotesFactory

from umlio.XMLConstants import XmlConstants

from umlio.deserializer.XmlToPyut import XmlToPyut


class XmlNotesToUmlNotes:
    def __init__(self):
        self.logger: Logger = getLogger(__name__)

        self._xmlToPyut: XmlToPyut = XmlToPyut()

    def deserialize(self, umlDiagramElement: Element) -> UmlNotes:
        """

        Args:
            umlDiagramElement:  The Element document

        Returns:  deserialized UmlNote objects if any exist, else an empty list
        """
        umlNotes:     UmlNotes = umlNotesFactory()
        noteElements: Elements = cast(Elements, umlDiagramElement.get_elements(XmlConstants.ELEMENT_UML_NOTE))

        for noteElement in noteElements:
            self.logger.debug(f'{noteElement}')

            graphicInformation: GraphicInformation = GraphicInformation.toGraphicInfo(graphicElement=noteElement)
            pyutNote:           PyutNote           = self._xmlToPyut.noteToPyutNote(umlNoteElement=noteElement)
            umlNote:            UmlNote            = UmlNote(pyutNote=pyutNote)

            umlNote.id       = graphicInformation.id
            umlNote.size     = graphicInformation.size
            umlNote.position = graphicInformation.position

            umlNotes.append(umlNote)

        return umlNotes




from logging import Logger
from logging import getLogger
from pathlib import Path

from zlib import compress
from zlib import ZLIB_VERSION

from umlio.IOTypes import PROJECT_SUFFIX
from umlio.IOTypes import UmlProject
from umlio.IOTypes import XML_SUFFIX
from umlio.serializer.UmlShapesToXml import UmlShapesToXml


class Writer:
    """
    A shim on top of the UML serialization layer;
    The write only writes the latest XML version
    """

    def __init__(self):

        self.logger: Logger = getLogger(__name__)

    def writeFile(self, umlProject: UmlProject, fileName: Path):
        """
        Writes to a compressed project file file

        Args:
            umlProject: The project we have to serialize
            fileName:   Where to write the XML;  Should be a full qualified file name
        """
        if fileName.suffix != PROJECT_SUFFIX:
            fileName = fileName / PROJECT_SUFFIX

        umlShapesToXml: UmlShapesToXml = UmlShapesToXml(projectFileName=fileName, projectCodePath=umlProject.codePath)

        for umlDiagram in umlProject.umlDocuments.values():
            umlShapesToXml.serialize(umlDiagram=umlDiagram)

        rawXml: str = umlShapesToXml.xml

        self.logger.debug(f'{ZLIB_VERSION=}')
        byteText:        bytes  = rawXml.encode()
        compressedBytes: bytes = compress(byteText)

        fileName.write_bytes(compressedBytes)

    def writeXmlFile(self, umlProject: UmlProject, fileName: Path):
        """
        Writes to an XML file
        Args:
            umlProject:   The project we have to serialize
            fileName:     Where to write the XML;  Should be a full qualified file name
        """
        if fileName.suffix != XML_SUFFIX:
            fileName = fileName / XML_SUFFIX

        umlToXml: UmlShapesToXml = UmlShapesToXml(projectFileName=umlProject.fileName, projectCodePath=umlProject.codePath)

        for umlDiagram in umlProject.umlDocuments.values():
            umlToXml.serialize(umlDiagram=umlDiagram)

        umlToXml.writeXml(fileName=fileName)




from pathlib import Path

from logging import Logger
from logging import getLogger

from zlib import decompress
from zlib import ZLIB_VERSION

from umlio.IOTypes import PROJECT_SUFFIX
from umlio.IOTypes import UmlProject
from umlio.IOTypes import XML_SUFFIX
from umlio.deserializer.XmlToUmlShapes import XmlToUmlShapes
from umlio.exceptions.UnsupportedFileTypeException import UnsupportedFileTypeException


class Reader:
    """
    """
    def __init__(self):

        self.logger: Logger = getLogger(__name__)

    def readProjectFile(self, fileName: Path) -> UmlProject:
        """
        Parse the input PROJECT_SUFFIX file

        Args:
            fileName: The fully qualified file name
        """
        suffix: str = fileName.suffix
        if len(suffix) < 0 or suffix != PROJECT_SUFFIX:
            raise UnsupportedFileTypeException(message=f'File does not end with {PROJECT_SUFFIX} suffix')

        rawXmlString: str = self._decompressFile(fileName=fileName)

        xmlToUmlShapes: XmlToUmlShapes = XmlToUmlShapes()

        xmlToUmlShapes.deserializeXml(xmlString=rawXmlString, fileName=fileName)

        return xmlToUmlShapes.umlProject

    def readXmlFile(self, fileName: Path) -> UmlProject:
        """
        Parse the input XML file;

        Args:
            fileName: Fully qualified file name
        """
        suffix: str = fileName.suffix
        if len(suffix) < 0 or suffix != XML_SUFFIX:
            raise UnsupportedFileTypeException(message=f'File does not end with .xml suffix')

        xmlToUmlShapes: XmlToUmlShapes = XmlToUmlShapes()

        xmlToUmlShapes.deserializeXmlFile(fileName=fileName)

        return xmlToUmlShapes.umlProject

    def _decompressFile(self, fileName: Path) -> str:
        """
        Decompresses a previously Pyut compressed file

        Args:
            fileName: Fully qualified file name with a PROJECT_SUFFIX suffix

        Returns:  A raw XML String
        """
        try:
            with open(fileName, "rb") as compressedFile:
                compressedData: bytes = compressedFile.read()
        except (ValueError, Exception) as e:
            self.logger.error(f'decompress open:  {e}')
            raise e
        else:
            self.logger.info(f'{ZLIB_VERSION=}')
            xmlBytes:  bytes = decompress(compressedData)  # has b`....` around it
            xmlString: str   = xmlBytes.decode()
            self.logger.debug(f'Document read:\n{xmlString}')

        return xmlString




from typing import cast

from logging import Logger
from logging import getLogger

from xml.etree.ElementTree import Element
from xml.etree.ElementTree import SubElement

from umlshapes.shapes.UmlNote import UmlNote

from umlio.IOTypes import UmlNotes

from umlio.serializer.BaseUmlToXml import BaseUmlToXml
from umlio.serializer.PyutToXml import PyutToXml
from umlio.XMLConstants import XmlConstants


class UmlNotesToXml(BaseUmlToXml):
    def __init__(self):
        super().__init__()
        self.logger: Logger = getLogger(__name__)

        self._pyutToXml: PyutToXml = PyutToXml()

    def serialize(self, documentTop: Element, umlNotes: UmlNotes) -> Element:

        for note in umlNotes:
            umlNote: UmlNote = cast (UmlNote, note)
            umlNoteElement: Element = self._umlNoteToXml(documentTop=documentTop, umlNote=umlNote)
            self._pyutToXml.pyutNoteToXml(pyutNote=umlNote.pyutNote, umlNoteElement=umlNoteElement)

        return documentTop

    def _umlNoteToXml(self, documentTop: Element, umlNote: UmlNote) -> Element:

        attributes = self._umlBaseAttributes(umlShape=umlNote)
        umlNoteSubElement: Element = SubElement(documentTop, XmlConstants.ELEMENT_UML_NOTE, attrib=attributes)

        return umlNoteSubElement




from typing import cast

from logging import Logger
from logging import getLogger

from xml.etree.ElementTree import Element
from xml.etree.ElementTree import SubElement

from umlshapes.shapes.UmlText import UmlText

from umlio.IOTypes import UmlTexts

from umlio.serializer.BaseUmlToXml import BaseUmlToXml
from umlio.serializer.PyutToXml import PyutToXml
from umlio.XMLConstants import XmlConstants


class UmlTextsToXml(BaseUmlToXml):
    def __init__(self):
        super().__init__()
        self.logger: Logger = getLogger(__name__)

        self._pyutToXml: PyutToXml = PyutToXml()

    def serialize(self, documentTop: Element, umlTexts: UmlTexts) -> Element:

        for text in umlTexts:
            umlText: UmlText = cast(UmlText, text)
            umlTextElement: Element = self._umlTextToXml(documentTop=documentTop, umlText=umlText)
            self._pyutToXml.pyutTextToXml(pyutText=umlText.pyutText, umlTextElement=umlTextElement)

        return documentTop

    def _umlTextToXml(self, documentTop: Element, umlText: UmlText) -> Element:

        attributes = self._umlBaseAttributes(umlShape=umlText)
        umlTextSubElement: Element = SubElement(documentTop, XmlConstants.ELEMENT_UML_TEXT, attrib=attributes)

        return umlTextSubElement




from typing import cast

from logging import Logger
from logging import getLogger

from untangle import Element

from pyutmodelv2.PyutText import PyutText

from umlshapes.shapes.UmlText import UmlText

from umlio.IOTypes import Elements
from umlio.IOTypes import GraphicInformation
from umlio.IOTypes import UmlTexts
from umlio.IOTypes import umlTextsFactory

from umlio.deserializer.XmlToPyut import XmlToPyut

from umlio.XMLConstants import XmlConstants


class XmlTextsToUmlTexts:
    """
    Yes, I know bad English
    """
    def __init__(self):

        super().__init__()

        self.logger: Logger = getLogger(__name__)

        self._xmlToPyut: XmlToPyut = XmlToPyut()

    def deserialize(self, umlDiagramElement: Element) -> UmlTexts:
        """

        Args:
            umlDiagramElement:  The Element document

        Returns:  deserialized UmlText objects if any exist, else an empty list
        """

        umlTexts:     UmlTexts = umlTextsFactory()
        textElements: Elements = cast(Elements, umlDiagramElement.get_elements(XmlConstants.ELEMENT_UML_TEXT))

        for textElement in textElements:
            self.logger.debug(f'{textElement}')

            graphicInformation: GraphicInformation = GraphicInformation.toGraphicInfo(graphicElement=textElement)
            pyutText:           PyutText           = self._xmlToPyut.textToPyutText(umlTextElement=textElement)
            umlText:            UmlText            = UmlText(pyutText=pyutText)

            umlText.id       = graphicInformation.id
            umlText.size     = graphicInformation.size
            umlText.position = graphicInformation.position

            umlTexts.append(umlText)

        return umlTexts
